﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text.RegularExpressions;

namespace Inedo.Consolation
{
    /// <summary>
    /// Contains methods for initializing a standard console application.
    /// </summary>
    public static class ConsoleApplication
    {
        /// <summary>
        /// Runs the type which contains the application entry point as a console application.
        /// </summary>
        /// <returns>Process exit code.</returns>
        /// <exception cref="InvalidOperationException">Entry point could not be determined.</exception>
        public static int Run()
        {
            return Run((string[])null);
        }
        /// <summary>
        /// Runs the type which contains the application entry point as a console application.
        /// </summary>
        /// <param name="args">Console input arguments.</param>
        /// <returns>Process exit code.</returns>
        /// <exception cref="InvalidOperationException">Entry point could not be determined.</exception>
        public static int Run(string[] args)
        {
            var entryAssembly = Assembly.GetEntryAssembly();
            if (entryAssembly == null)
                throw new InvalidOperationException("Could not determine entry assembly.");

            var entryPoint = entryAssembly.EntryPoint;
            if (entryPoint == null)
                throw new InvalidOperationException("Could not determine entry point.");

            return Run(entryPoint.DeclaringType, args ?? (Environment.GetCommandLineArgs().Skip(1).ToArray()));
        }
        /// <summary>
        /// Runs the specified type as a console application.
        /// </summary>
        /// <param name="applicationType">The type that defines a console application.</param>
        /// <param name="args">Console input arguments.</param>
        /// <returns>Process exit code.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="applicationType"/> or <paramref name="args"/> is null.</exception>
        public static int Run(Type applicationType, string[] args)
        {
            if (applicationType == null)
                throw new ArgumentNullException(nameof(applicationType));
            if (args == null)
                throw new ArgumentNullException(nameof(args));

            if (args.Length == 1 && string.Equals(args[0], "help", StringComparison.OrdinalIgnoreCase))
            {
                WriteUsage(applicationType);
                return 0;
            }

            if (args.Length >= 2 && string.Equals(args[0], "help", StringComparison.OrdinalIgnoreCase))
            {
                var helpCommand = GetBestMatchingCommand(GetCommands(applicationType), args[1]);
                if (helpCommand != null)
                {
                    WriteCommandUsage(applicationType.Assembly.GetName(), helpCommand);
                    return 0;
                }
                else
                {
                    Console.Error.WriteLine("Invalid command: " + args[1]);
                    Console.WriteLine();
                    WriteUsage(applicationType);
                    return -1;
                }
            }

            if (args.Length >= 1 && string.Equals(args[0], "interactive", StringComparison.OrdinalIgnoreCase))
            {
                RunInteractive(applicationType);
                return 0;
            }

            try
            {
                var command = ProcessArguments(applicationType, args);
                if (command == null)
                {
                    WriteUsage(applicationType);
                    return 0;
                }

                return command.Invoke();
            }
            catch (ConsoleArgumentException ex)
            {
                Console.Error.WriteLine(ex.Message);
                Console.WriteLine();
                WriteUsage(applicationType);
                return -1;
            }
            catch (TargetInvocationException ex)
            {
                var inner = ex.InnerException;
                Console.Error.WriteLine("Unhandled exception: " + inner.ToString());
                return -2;
            }
        }

        private static void RunInteractive(Type type)
        {
            string input;
            do
            {
                var globalOptions = GetGlobalOptions(type).ToList();
                var commands = GetCommands(type).ToList();

                var assemblyName = type.Assembly.GetName();
                Console.WriteLine();
                Console.WriteLine("{0} {1}", assemblyName.Name, assemblyName.Version);
                Console.WriteLine("Interactive Mode");
                Console.WriteLine();

                int index = 1;
                foreach (var command in commands)
                {
                    Console.WriteLine("[{0}]: {1}", index++, command.Name);
                    if (command.Description != null)
                        Console.WriteLine("\t{0}", command.Description);
                }

                if (globalOptions.Count > 0)
                    Console.WriteLine("[o]: Configure global options");

                Console.WriteLine("[q]: Exit");
                Console.WriteLine();

                Console.Write(">> ");
                input = Console.ReadLine() ?? "q";

                if (string.Equals(input, "o", StringComparison.OrdinalIgnoreCase))
                    InteractiveOptionMenu(globalOptions);
                else
                {
                    int parsedIndex;
                    if (int.TryParse(input, out parsedIndex) && parsedIndex >= 1 && parsedIndex <= commands.Count)
                    {
                        var command = commands[parsedIndex - 1];
                        if (command.Options.Count == 0)
                            InvokeCommandInteractive(command);
                        else
                            InteractiveCommandMenu(assemblyName, command);
                    }
                }
            }
            while (!string.Equals(input, "q", StringComparison.OrdinalIgnoreCase));
        }
        private static void InteractiveCommandMenu(AssemblyName assemblyName, ConsoleCommand command)
        {
            string input;
            do
            {
                Console.WriteLine();
                Console.WriteLine("{0} {1}", assemblyName.Name, command.Name);
                Console.WriteLine();

                if (command.Options.Count > 0)
                {
                    int index = 1;
                    foreach (var option in command.Options)
                    {
                        if (option.IsBrowsable)
                        {
                            Console.WriteLine("[{0}]: {1}={2}", index++, option.Name, option.CurrentValue ?? option.DefaultValue ?? "<not set>");
                            if (option.Description != null)
                                Console.WriteLine("\t{0}", option.Description);
                        }
                        else
                        {
                            index++;
                        }
                    }
                }

                Console.WriteLine("[r]: Run command");
                Console.WriteLine("[q]: Exit");

                Console.WriteLine();
                Console.Write(">> ");
                input = (Console.ReadLine() ?? "q").Trim();

                if (string.Equals(input, "r", StringComparison.OrdinalIgnoreCase))
                {
                    InvokeCommandInteractive(command);
                    return;
                }

                int parsedIndex;
                if (int.TryParse(input, out parsedIndex) && parsedIndex >= 1 && parsedIndex <= command.Options.Count)
                {
                    var option = command.Options[parsedIndex - 1];
                    if (option.IsFlag)
                        option.SetFlag(!(option.CurrentValue as bool? ?? false));
                    else
                        InteractiveOptionPrompt(option);
                }
            }
            while (!string.Equals(input, "q", StringComparison.OrdinalIgnoreCase));
        }
        private static void InteractiveOptionMenu(List<GlobalConsoleOption> options)
        {
            string input;
            do
            {
                int index = 1;
                foreach (var option in options)
                {
                    Console.WriteLine("[{0}]: {1}={2}", index++, option.Name, option.CurrentValue);
                    if (option.Description != null)
                        Console.WriteLine("\t{0}", option.Description);
                }

                Console.WriteLine();
                input = (Console.ReadLine() ?? "q").Trim();

                int parsedIndex;
                if (int.TryParse(input, out parsedIndex) && parsedIndex >= 1 && parsedIndex <= options.Count)
                {
                    var option = options[parsedIndex - 1];
                    if (option.IsFlag)
                        option.SetFlag(!(option.CurrentValue as bool? ?? false));
                    else
                        InteractiveOptionPrompt(option);
                }
            }
            while (!string.Equals(input, "q", StringComparison.OrdinalIgnoreCase));
        }
        private static void InteractiveOptionPrompt(ConsoleOption option)
        {
            Console.WriteLine("Enter a value for {0}:", option.Name);
            if (option.DefaultValue != null || option.CurrentValue != null)
                Console.Write("[{0}] >> ", option.CurrentValue ?? option.DefaultValue);
            else
                Console.Write(">> ");

            var input = (Console.ReadLine() ?? string.Empty).Trim();
            if (!string.IsNullOrEmpty(input))
            {
                try
                {
                    option.SetValue(input);
                }
                catch (ConsoleArgumentException ex)
                {
                    Console.Error.WriteLine(ex.Message);
                }
            }
        }
        private static void InvokeCommandInteractive(ConsoleCommand command)
        {
            Console.WriteLine();
            try
            {
                command.Invoke();
            }
            catch (ConsoleArgumentException ex)
            {
                Console.Error.WriteLine(ex.Message);
            }
            catch (TargetInvocationException ex)
            {
                var inner = ex.InnerException;
                Console.Error.WriteLine("Unhandled exception: " + inner.ToString());
            }
        }

        private static ConsoleCommand ProcessArguments(Type type, string[] args)
        {
            var parsedArgs = args
                .Select(ParseArgument)
                .ToList();

            var globalOptions = GetGlobalOptions(type).ToList();
            int globalOptionCount = 0;
            foreach (var arg in parsedArgs.TakeWhile(a => a.IsOption))
            {
                var option = GetBestMatchingOption(globalOptions, arg.Name);
                if (option == null)
                    throw new ConsoleArgumentException("Invalid option: " + arg.RawValue);

                if (option.IsFlag && arg.OptionValue != null)
                    throw new ConsoleArgumentException(string.Format("Value ={0} is unexpected after flag option {1}", arg.OptionValue, arg.Name));

                if (!option.IsFlag && arg.OptionValue == null)
                    throw new ConsoleArgumentException(string.Format("Value (=[value]) is required after option {0}", arg.Name));

                if (option.IsFlag)
                    option.SetFlag(arg.EnableDisable ?? true);
                else
                    option.SetValue(arg.OptionValue);

                globalOptionCount++;
            }

            if (globalOptionCount >= parsedArgs.Count)
                return null;

            var commandName = parsedArgs[globalOptionCount].Name;
            var command = GetBestMatchingCommand(GetCommands(type), commandName);
            if (command == null)
                throw new ConsoleArgumentException("Invalid command: " + commandName);

            int positionalIndex = 0;
            foreach (var arg in parsedArgs.Skip(globalOptionCount + 1))
            {
                CommandConsoleOption option;
                if (arg.IsOption)
                {
                    option = GetBestMatchingOption(command.Options, arg.Name);
                    if (option == null)
                        throw new ConsoleArgumentException("Invalid option: " + arg.RawValue);
                }
                else
                {
                    option = command.GetPositionalOption(positionalIndex);
                    if (option == null)
                        throw new ConsoleArgumentException("Invalid option: " + arg.RawValue);

                    positionalIndex++;
                }

                if (option.IsFlag)
                    option.SetFlag(arg.EnableDisable ?? true);
                else
                    option.SetValue(arg.OptionValue ?? arg.Name);
            }

            return command;
        }
        private static IEnumerable<GlobalConsoleOption> GetGlobalOptions(Type type)
        {
            return from p in type.GetProperties()
                   where Attribute.IsDefined(p, typeof(ConsoleOptionAttribute)) && p.CanWrite && p.GetSetMethod().IsStatic
                   select new GlobalConsoleOption(p);
        }
        private static IEnumerable<ConsoleCommand> GetCommands(Type type)
        {
            return from m in type.GetMethods()
                   where Attribute.IsDefined(m, typeof(ConsoleCommandAttribute)) && m.IsStatic
                   select new ConsoleCommand(m);
        }
        private static ProcessedArgument ParseArgument(string arg)
        {
            var optionMatch = Regex.Match(arg, @"^-{1,2}(?<1>.+)$", RegexOptions.ExplicitCapture);
            if (optionMatch.Success)
            {
                var optionText = optionMatch.Groups[1].Value;
                var keyValueMatch = Regex.Match(optionText, @"^(?<1>[^=]+)=(?<2>.*)$", RegexOptions.ExplicitCapture);
                if (keyValueMatch.Success)
                {
                    return new ProcessedArgument
                    {
                        IsOption = true,
                        Name = keyValueMatch.Groups[1].Value,
                        OptionValue = keyValueMatch.Groups[2].Value,
                        RawValue = arg
                    };
                }
                else
                {
                    bool? inclused = optionText.EndsWith("+") ? true
                        : optionText.EndsWith("-") ? (bool?)false
                        : null;

                    return new ProcessedArgument
                    {
                        IsOption = true,
                        Name = optionText.TrimEnd('-', '+'),
                        EnableDisable = inclused,
                        RawValue = arg
                    };
                }
            }

            return new ProcessedArgument
            {
                Name = arg,
                RawValue = arg
            };
        }
        private static void WriteUsage(Type type)
        {
            var globalOptions = GetGlobalOptions(type).ToList();
            var assemblyName = type.Assembly.GetName();
            Console.WriteLine("{0} {1}", assemblyName.Name, assemblyName.Version);
            if (globalOptions.Count > 0)
                Console.WriteLine($"Usage: {assemblyName.Name} [options] [command]");
            else
                Console.WriteLine($"Usage: {assemblyName.Name} [command]");

            Console.WriteLine();

            if (globalOptions.Count > 0)
            {
                Console.WriteLine("Options:");
                foreach (var option in globalOptions)
                {
                    var rubbish = option.Name;
                    if (!option.IsFlag)
                        rubbish = rubbish + "=[value]";

                    var name = string.Join(", ", option.Aliases.Concat(new[] { "--" + rubbish }));
                    Console.Write(name);

                    if (option.Description != null)
                    {
                        Console.Write("\t");
                        Console.Write(option.Description);
                    }

                    Console.WriteLine();
                }
            }

            Console.WriteLine("Commands:");
            foreach (var command in GetCommands(type))
            {
                if (command.IsBrowsable)
                {
                    var name = string.Join(", ", command.Aliases.Concat(new[] { command.Name }));
                    Console.Write(name);

                    if (command.Description != null)
                    {
                        Console.Write("\t");
                        Console.Write(command.Description);
                    }

                    Console.WriteLine();
                }
            }

            Console.WriteLine();
            Console.WriteLine("For help on a specific command, type: {0} help [command]", assemblyName.Name);
            Console.WriteLine("To run interactively, type: {0} interactive", assemblyName.Name);
        }
        private static void WriteCommandUsage(AssemblyName assemblyName, ConsoleCommand command)
        {
            var visibleOptions = command.Options.Where(o => o.IsBrowsable).ToList();

            var positionalArgs = from a in visibleOptions
                                 where a.IsPositional
                                 select a;

            bool hasNonPositionalArgs = visibleOptions.Any(a => !a.IsPositional);

            if (command.Description != null)
            {
                Console.WriteLine(command.Description);
                Console.WriteLine();
            }

            Console.Write("Usage: {0} {1}", assemblyName.Name, command.Name);
            if (positionalArgs.Any())
                Console.Write(" " + string.Join(" ", positionalArgs.Select(a => "[" + a.Name + "]")));
            if (hasNonPositionalArgs)
                Console.Write(" [options]");

            if (visibleOptions.Count > 0)
            {
                Console.WriteLine();
                Console.WriteLine("Arguments:");
                foreach (var arg in visibleOptions)
                {
                    var rubbish = arg.Name;
                    if (!arg.IsFlag)
                        rubbish = rubbish + "=[value]";

                    var name = string.Join(", ", arg.Aliases.Concat(new[] { "--" + rubbish }));
                    Console.Write(name);

                    if (arg.Description != null)
                    {
                        Console.Write("\t");
                        Console.Write(arg.Description);
                    }

                    Console.WriteLine();
                }
            }

            Console.WriteLine();
        }
        private static TOption GetBestMatchingOption<TOption>(IEnumerable<TOption> options, string optionName)
            where TOption : ConsoleOption
        {
            var matches = from o in options
                           where string.Equals(o.Name, optionName, StringComparison.OrdinalIgnoreCase) || o.Aliases.Contains(optionName, StringComparer.OrdinalIgnoreCase)
                           orderby o.Name == optionName descending, string.Equals(o.Name, optionName, StringComparison.OrdinalIgnoreCase) descending, o.Aliases.Contains(optionName) descending
                           select o;

            return matches.FirstOrDefault();
        }
        private static ConsoleCommand GetBestMatchingCommand(IEnumerable<ConsoleCommand> commands, string commandName)
        {
            var matches = from c in commands
                          where string.Equals(c.Name, commandName, StringComparison.OrdinalIgnoreCase) || c.Aliases.Contains(commandName, StringComparer.OrdinalIgnoreCase)
                          orderby c.Name == commandName descending, string.Equals(c.Name, commandName, StringComparison.OrdinalIgnoreCase) descending, c.Aliases.Contains(commandName) descending
                          select c;

            return matches.FirstOrDefault();
        }
    }
}
