﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Inedo.Agents.Commands;
using Inedo.Agents.Hosting;

namespace Inedo.Agents.Service
{
    internal sealed class InedoAgentServer : IDisposable
    {
        private static readonly Lazy<AgentCommandDispatcher> CommandDispatcher = new Lazy<AgentCommandDispatcher>(() => new AgentCommandDispatcher(typeof(UpdateAgentCommand).Assembly), LazyThreadSafetyMode.PublicationOnly);
        private Dictionary<int, ClientConnection> clients = new Dictionary<int, ClientConnection>();
        private Dictionary<int, HostedProcess> hostedProcesses = new Dictionary<int, HostedProcess>();
        private ConcurrentDictionary<int, OutgoingMessageQueue> outgoingMessagesBySession = new ConcurrentDictionary<int, OutgoingMessageQueue>();
        private TcpListener tcpListener;

        public InedoAgentServer(IPAddress address, int port)
        {
            this.tcpListener = new TcpListener(address, port);
            HostedProcessManager.Instance.ResponseReceived += this.HostedProcess_ResponseReceived;
        }

        public bool IsStarted { get; private set; }

        public async Task WaitForConnectionsAsync()
        {
            if (this.IsStarted)
                throw new InvalidOperationException("Already started.");

            this.IsStarted = true;

            Log.Info("Listening for connections...");
            tcpListener.Start();

            while (this.IsStarted)
            {
                try
                {
                    var channel = await this.AcceptConnectionAsync().ConfigureAwait(false);
                    Log.Info("Connection request from " + channel.TcpEndPoint);

                    int requestedSessionId;
                    try
                    {
                        requestedSessionId = await channel.ReceiveHandshake().ConfigureAwait(false);
                    }
                    catch (Exception ex)
                    {
                        Log.Error($"Bad handshake from {channel.TcpEndPoint}: " + ex.ToString());
                        try { channel.Dispose(); }
                        catch { }
                        continue;
                    }

                    int sessionId;
                    lock (this.clients)
                    {
                        channel.Disconnected += this.Channel_Disconnected;

                        sessionId = this.GetSessionId(requestedSessionId);
                        this.clients.Add(sessionId, channel);
                    }

                    try
                    {
                        await channel.SendHandshakeResponse(sessionId).ConfigureAwait(false);
                    }
                    catch (Exception ex)
                    {
                        Log.Error($"Error sending handshake response to {channel.TcpEndPoint}: " + ex.ToString());
                        try { channel.Dispose(); }
                        catch { }
                        continue;
                    }

                    Log.Info("Connection established with " + channel.TcpEndPoint);
                    this.ProcessIncomingMessagesAsync(sessionId, channel);
                }
                catch (OperationCanceledException)
                {
                    this.IsStarted = false;
                    return;
                }
                catch
                {
                    this.IsStarted = false;
                    throw;
                }
            }
        }

        private async void ProcessIncomingMessagesAsync(int sessionId, ClientConnection channel)
        {
            await Task.Yield();

            // if this is a reconnection, try to flush the queue
            OutgoingMessageQueue queue;
            if (this.outgoingMessagesBySession.TryGetValue(sessionId, out queue))
                queue.Flush();

            while (channel.IsConnected)
            {
                AgentMessage message;
                try
                {
                    message = await channel.ReadMessageAsync().ConfigureAwait(false);
                }
                catch (ObjectDisposedException)
                {
                    message = null;
                }
                catch (Exception ex)
                {
                    // this is kind of rubbish
                    Log.Error($"Unhandled exception processing message from {channel.TcpEndPoint}: " + ex.ToString());
                    this.Channel_Disconnected(channel, EventArgs.Empty);
                    message = null;
                }

                if (message == null)
                    return;

                this.ProcessReceivedMessageInternal(sessionId, channel, message);
            }
        }
        private async void ProcessReceivedMessageInternal(int sessionId, ClientConnection channel, AgentMessage message)
        {
            await Task.Yield();

            Stream responseStream = null;

            try
            {
                if (message.AgentType == AgentType.Meta)
                {
                    // meta agent request
                    responseStream = await CommandDispatcher.Value.ExecuteCommandAsync(message.MessageTypeCode, message.MessageData).ConfigureAwait(false);
                }
                else
                {
                    // otter or bm agent request
                    await HostedProcessManager.Instance.GetVirtualHost(message.AgentType).SendMessageAsync(sessionId, message).ConfigureAwait(false);
                }
            }
            catch (Exception ex)
            {
                message.MessageData.Position = 0;
                if (message.MessageData.ReadByte() == 1)
                {
                    var id = new byte[16];
                    message.MessageData.Read(id, 0, 16);
                    responseStream = new MemoryStream(200);
                    var writer = new BinaryWriter(responseStream, new UTF8Encoding(false));
                    writer.Write(id);
                    writer.Write((byte)1);
                    AgentCommandDispatcher.SerializeException(ex, writer);
                }
            }

            if (responseStream != null)
            {
                responseStream.Position = 0;
                this.PostMessage(sessionId, new OutgoingMessage(MessageTypeCodes.MessageResponse, responseStream));
            }
        }

        public void Dispose()
        {
            try { this.tcpListener.Stop(); }
            catch { }

            ClientConnection[] clients;
            OutgoingMessageQueue[] queues;
            lock (this.clients)
            {
                clients = this.clients.Values.ToArray();
                this.clients.Clear();

                queues = this.outgoingMessagesBySession.Values.ToArray();
                this.outgoingMessagesBySession.Clear();
            }

            foreach (var client in clients)
            {
                try { client.Dispose(); }
                catch { }
            }

            foreach (var queue in queues)
            {
                try { queue.Dispose(); }
                catch { }
            }
        }

        private void Channel_Disconnected(object sender, EventArgs e)
        {
            var client = (ClientConnection)sender;
            Log.Info(client.TcpEndPoint + " disconnected.");
            lock (this.clients)
            {
                int? sessionId = null;
                client.Disconnected -= this.Channel_Disconnected;
                foreach (var item in this.clients)
                {
                    if (item.Value == client)
                    {
                        this.clients.Remove(item.Key);
                        sessionId = item.Key;
                        break;
                    }
                }

                if (sessionId != null)
                {
                    OutgoingMessageQueue queue;
                    if (this.outgoingMessagesBySession.TryGetValue(sessionId.Value, out queue))
                    {
                        if (queue.IsEmpty)
                        {
                            try { queue.Dispose(); }
                            catch { }
                            this.outgoingMessagesBySession.TryRemove(sessionId.Value, out queue);
                        }
                    }

                    HostedProcessManager.Instance.SessionDisconnected(sessionId.Value);
                }
            }

            // ensure connection gets closed
            try { client.Dispose(); }
            catch { }
        }

        private int GetSessionId(int requestedId)
        {
            bool useRequested = !this.clients.ContainsKey(requestedId) && this.outgoingMessagesBySession.ContainsKey(requestedId);
            if (useRequested)
                return requestedId;

            using (var random = RandomNumberGenerator.Create())
            {
                var bytes = new byte[4];
                int id;
                do
                {
                    random.GetBytes(bytes);
                    id = BitConverter.ToInt32(bytes, 0);
                }
                while (this.clients.ContainsKey(id));

                return id;
            }
        }
        private void PostMessage(int sessionId, OutgoingMessage message)
        {
            var queue = this.outgoingMessagesBySession.GetOrAdd(sessionId, this.CreateMessageQueue);
            queue.PostMessage(message);
        }

        private OutgoingMessageQueue CreateMessageQueue(int sessionId)
        {
            return new OutgoingMessageQueue(m => this.SendMessageAsync(sessionId, m));
        }
        private Task SendMessageAsync(int sessionId, OutgoingMessage message)
        {
            ClientConnection channel;
            lock (this.clients)
            {
                if (!this.clients.TryGetValue(sessionId, out channel))
                    throw new Exception();
            }

            message.MessageData.Position = 0;
            return channel.SendMessageAsync(message.MessageTypeCode, message.MessageData);
        }

        private Task<ClientConnection> AcceptConnectionAsync()
        {
            var completionSource = new TaskCompletionSource<ClientConnection>();
            this.tcpListener.BeginAcceptTcpClient(this.HandleEndAcceptTcpClient, completionSource);
            return completionSource.Task;
        }

        private void HandleEndAcceptTcpClient(IAsyncResult result)
        {
            var completionSource = (TaskCompletionSource<ClientConnection>)result.AsyncState;
            try
            {
                var tcpClient = this.tcpListener.EndAcceptTcpClient(result);
                var channel = new ClientConnection(tcpClient.GetStream());
                channel.TcpEndPoint = tcpClient.Client.RemoteEndPoint;
                completionSource.TrySetResult(channel);
            }
            catch (ObjectDisposedException)
            {
                completionSource.TrySetCanceled();
            }
            catch (Exception ex)
            {
                completionSource.TrySetException(ex);
            }
        }
        private void HostedProcess_ResponseReceived(object sender, CommandResponseEventArgs e)
        {
            this.PostMessage(e.SessionId, new OutgoingMessage(e.Process.AgentType, e.MessageId, e.Data));
        }
    }
}
